#
# "Programmer's Life" Simulation
# Gbor SEBESTYN <segabor@elte.hu>
#


require 'fstate.rb'
require 'timer.rb'

#
# GLOBAL VARS
#

$DEV_CNT=rand(6)
$tm = Timer.new
$actors = []


#
# HELPER MODULES AND CLASSES
#

module NameGenerator
    LNAMES = [ "Kovacs", "Lakatos", "Szabo", "Horvat", "Lovas", "Gesztes", "Szekely", "Katona", "Karasz", "Hetenyi", "Aradi" ]
    FNAMES_M = [ "Karoly", "Peter", "Geza", "Istvan", "Jozsef", "Artur", "Laszlo", "Gabor", "Szabolcs", "Miklos", "Menyhert", "Elemer", "Balazs", "Vilmos", "Almos" ]
    FNAMES_F = [ "Szilvi", "Katalin", "Jolan", "Maria", "Irma", "Kriszta", "Edina", "Mariann", "Andrea", "Adrienn", "Tunde", "Zsuzsa" ]
    def NameGenerator::GetName(f = true)
        srand(Time.now.usec)
        s1 = LNAMES[rand(LNAMES.size*2)/2]
        # LNAMES[rand(LNAMES.size)] + " " + ( f == true ? FNAMES_M[rand(FNAMES_M.size)] : FNAMES_F[rand(FNAMES_F.size)] )
        s2 = ( f == true ? FNAMES_M[rand(FNAMES_M.size*2)/2] : FNAMES_F[rand(FNAMES_F.size*2)/2] )
        s1 + " " + s2
    end
end


class Actor
	attr	:is_alive

	def initialize
		@is_alive = true
	end

	def is_alive?
		@is_alive
	end
	
	def kill
		@alive = false
		finish
	end

	def finish
	end

    def do_it
    end
end


#
# MUNKAHELY
#
class Munkahely < Actor
    attr    :alkalmazottak
    attr	:fonok
    attr	:titkarno
    
    attr	:uzenet
    
    def initialize
        @alkalmazottak = []
        @uzenet = { "fax" => 0, "emil" => 0, "telefon" => 0}
    end

    def fonok=(f)
        @fonok=f
    end

    def titkarno=(t)
        @titkarno=t
    end

    def +(a)
        @alkalmazottak << a unless a.is_a? Fonok
        self
    end

    def -(a)
        @alkalmazottak.delete(a)
        self
    end

	def canGoForLunch?(who)
		k = 0
		unless who.is_a? Fonok
			@alkalmazottak.each do |a|
				if a.havingLunch?
					k = k+1
				end
			end
		end
		k == 0
	end

	# main loop
	def do_it
		if @alkalmazottak.size <= 1
			$DEV_CNT.times {|i|
				f = Fejleszto.new(self)
				$actors << f
			}
		else
			if rand < 0.2
				k = rand(100)
				if (k % 14) != 0
					if (k % 2) == 0
						uzenet["fax"] += 1
						puts "[Uj fax erkezett!]"
					elsif (k % 4) == 0
						uzenet["telefon"] += 1
						puts "[Uj telefonhivas erkezett!]"
					else
						uzenet["emil"] += 1
						puts "[Uj emil erkezett!]"
					end
				end
			end
	
			# hiring new developer		
			if @alkalmazottak.size < 8 and rand < 0.05
				f = Fejleszto.new(self)
				$actors << f
			end
		end

	end
end


#
# DOLGOZO
#
class Dolgozo < Actor
    SHR   = 8      # 8:00 am
    THR   = 17     # 5:00 pm
    # personal identity
    attr    :gender
    attr    :name
    attr    :company
    # states
    attr    :loc_state

    def initialize(mh, k=-1)
    	super()
    
        mh += self
        @company    = mh   unless mh.nil?
    
        if k == -1
        	k = rand(10000)
	       	k = k % 2
        end
        @gender = ( k == 0 ? true : false)
        @name = NameGenerator::GetName( @gender )
    
        s = [ "s", "home", "work", "lunch", "away", "t" ]
        e = { "s" => [ "home", "work"], "home" => "work", "work" => [ "home", "lunch", "away" ], "lunch" => ["work", "home"], "away" => ["work", "home"] }
        @loc_state = FState.new(s, nil, nil, e)
        @loc_state.delegate = self

        # set start state
        # if $tm.dtime >= SHR*60 and $tm.dtime < THR*60
        #     @loc_state.setState("work")
        # else
        #    @loc_state.setState("home")
        # end
        @loc_state.setState("home")
    end

    def to_s
        @name + ", " + self.class.to_s + "[" + self.id.to_s + "]"
    end

	def company=(c)
		@company = c
	end

	def available?
		@loc_state.state == "work"
	end

	def working?
		@loc_state.state == "work"
	end

	def havingLunch?
		@loc_state.state == "lunch"
	end

    # location events
    def move_home_to_work
		puts "#{self} at work"
		# @loc_ev = $tm.adde(THR, 0, 0, self)
		@work_t_end = $tm.t + 8*3600 + rand(3600)
		@had_lunch = false
    end
    
    def enter_work
    end
    
    def enter_home
        puts "#{self} at home!"
        # @loc_ev = $tm.adde(SHR, 0, 0, self)
    end

    def enter_lunch
        lm = 10 + rand(15)
        puts "#{self} is out for lunch for #{lm} minutes!"
		@had_lunch = true
        @lunch_end = $tm.dtime + lm
        @lunch_back = @loc_state.last
    end

    def fire(t)
        if t == @loc_ev
            # location event arrived
            if @loc_state.state == "home"
                puts "#{self}: Let's go to work!"
                @loc_state.setState("work")
            else
                puts "#{self}: Go home!"
                @loc_state.setState("home")
            end
        end
    end

	# main lopp
    def do_it
		return if (not is_alive?) or @company.nil?
        case @loc_state.state
            when "home"
                if $tm.dtime >= SHR*60 and $tm.dtime < THR*60
                    @loc_state.setState("work")
                end
            when "work"
            	# $tm.dtime < SHR*60 or $tm.dtime >= THR*60
                if $tm.t > @work_t_end
                    @loc_state.setState("home")
                elsif (not @had_lunch) and @company.canGoForLunch?(self) and ($tm.dtime > 12*60)	#($tm.dtime - 12*60).abs < 30
                    @loc_state.setState("lunch")
                end
            when "lunch"
                if $tm.dtime >= @lunch_end
                    @loc_state.back
                end
        end
    end
    
    def finish
    	puts "#{self.to_s} bye-bye"
    end
end


#
# FONOK
#
class Fonok < Dolgozo
    def initialize(mh, k=-1)
        super(mh, k)
    end
    
    def disturb?
    	true
    end
    
    def got_message(key)
    	puts self.to_s + "got a #{key} message!"
    end
    
    def serve_cofee(count)
    	self.to_s + " got #{count} cup(s) of cofee"
    end
    
    def project_completed(who)
    end

	def do_it
		return if not is_alive?

		super
		
		if working?
			if rand < 0.05
				a = @company.alkalmazottak[rand(@company.alkalmazottak.size)]
				if a != @company.titkarno and not a.working?
					puts self.to_s + " fired #{a.to_s}"
					
					a.kill					
					a.company=nil
					
					@company.alkalmazottak.delete(a)
				end
			end
		end
	end
end


#
# ALKALMAZOTT
#
class Alkalmazott < Dolgozo
    def initialize(mh, k=-1)
        super(mh, k)
    end

    def got_message(key)
    	puts self.to_s + "got a #{key} message!"
    end
end


#
# TITKARNO
#
class Titkarno < Alkalmazott
    def initialize(mh, k=-1)
        super(mh, k)
    end
    
    def do_it
		return if not is_alive?

    	super
    	
    	case @loc_state.state
    		when "work"
    			if @company.fonok.disturb?
					# 1. check incoming messages (max 3 pieces)
					k = 3
					k = handle_incoming_msg("telefon", k)
					k = handle_incoming_msg("fax", k)
					k = handle_incoming_msg("emil", k)
					
					if rand < 0.2
						@company.fonok.serve_cofee 1
					end
				end
    	end
    end

	def handle_incoming_msg(key, m)
		if m > 0
			k = @company.uzenet[key]
			if k > 0
				if k > m
					h = m
					m = 0
					k -= m
				else
					h = k
					m -= k
					k = 0
				end
				puts self.to_s + "Handling #{h} of #{key} messages"
				@company.uzenet[key] = k
				h.times {|i|
					if rand > 0.2
						@company.fonok.got_message(key)
					else
						s = @company.alkalmazottak.size
						@company.alkalmazottak[rand(s)].got_message(key)
					end
				}
			end
		end
		m
	end
end

#
# FEJLESZTO
#
class Fejleszto < Alkalmazott
	attr	:has_work
    attr    :work_state
    attr    :work_t
	attr	:idle

    def initialize(mh, k=-1)
        super(mh, k)

        # 'dolgozik' substate
        s = [ "start", "tervez", "kodol", "support", "debuggol", "kesz" ]
        t = s[5]
        e = { "start" => "tervez", "tervez" => "kodol", "kodol" => ["support", "debuggol", "tervez"], "debuggol" => ["support", "kodol", "kesz"], "support" => ["kodol", "tervez"] }
        @has_work = false
        @idle = -1
        @work_state = FState.new(s, nil, t, e)
        @work_state.delegate = self
    end

	# work states
    
    def exit_start
		puts self.to_s + " is just starting a new project"
    end
    
    def enter_tervez
		puts self.to_s + " is planning"
    end

    def enter_kodol
		puts self.to_s + " is coding"
	end
	
    def enter_debuggol
		puts self.to_s + " is tracking bugs"
	end
	    
    def enter_support
		puts self.to_s + " calling support"
	end
	    

    def enter_kesz
		# notify boss!
		puts self.to_s + " is completed his/her latest project"
		@company.fonok.project_completed(self)
    end
    
    # main loop
    def do_it
		return if not is_alive?

    	super
    	
			if @has_work
				case @work_state.state
					when "start"
						@work_state.setState("tervez")
						@work_t = $tm.t + 3600 + rand(3600)
					when "tervez"
						if @work_t < $tm.t
							@work_state.setState("kodol")
							@work_t = $tm.t + 3600 + rand(3600)
						end
					when "kodol"
						if @work_t < $tm.t
							sts = @work_state.nextStates(@work_state.state)
							@work_state.setState(sts[rand(sts.size)])
							@work_t = $tm.t + 3600 + rand(3600)
						end
					when "support"
						if @work_t < $tm.t
							sts = @work_state.nextStates(@work_state.state)
							@work_state.setState(sts[rand(sts.size)])
							@work_t = $tm.t + 3600 + rand(3600)
						end
					when "debuggol"
						if @work_t < $tm.t
							sts = @work_state.nextStates(@work_state.state)
							@work_state.setState(sts[rand(sts.size)])
							@work_t = $tm.t + 3600 + rand(3600)
						end
					when "kesz"
						@has_work = false
				end
			else
				if rand < 0.3
					# has a new project!
					@has_work = true
					@work_state.reset
				else
					do_something
				end
			end
    end
    
    def do_something
		k = rand(3)
		if idle != k
			case k
				when 1
					puts self.to_s + " (Having Quake 3 Lan Party)"
				when 2
					puts self.to_s + " (Browsing cute girls on the net)"
				when 3
					puts self.to_s + " (Downloading some DiVX films)"
			end
		end
    end

	def working?
		@has_work
	end
end

mh1 = Munkahely.new
mh1.fonok = Fonok.new(mh1)
mh1.titkarno = Titkarno.new(mh1, 1)

$actors << mh1
$actors << mh1.fonok
$actors << mh1.titkarno

# main loop
puts
puts "The SiM begins at (#{$tm.t.hour}:#{$tm.t.min})"
puts

while true
	# make actors work
    $actors.each{|a| a.do_it}

    # leap timer a half hour
    $tm += 1800

	puts "(#{$tm.t.hour}:#{$tm.t.min})"

    # wait some ..
    sleep 1
end

